-- ViewportBorderImage
	ViewportBorderImage = ViewportBorderImage or class()

	function ViewportBorderImage:init(source, image_width, image_height, viewport_x, viewport_y, viewport_width, viewport_height)
		self._source = source

		self._width = image_width
		self._height = image_height

		self._viewport_x = viewport_x
		self._viewport_y = viewport_y

		self._viewport_width = viewport_width
		self._viewport_height = viewport_height

		self._position_x = 0
		self._position_y = 0

		self._scale_x = 1
		self._scale_y = 1

		self._death_wanted = false
	end

	function ViewportBorderImage:request_death()
		self._death_wanted = true
	end

	function ViewportBorderImage:update(t, dt, viewport_aspect_ratio)
		return self._death_wanted
	end

	function ViewportBorderImage:get_source()
		return self._source
	end

	function ViewportBorderImage:get_size()
		return {
			width = self._width,
			height = self._height
		}
	end

	function ViewportBorderImage:get_viewport_position()
		return {
			x = self._viewport_x,
			y = self._viewport_y
		}
	end

	function ViewportBorderImage:get_viewport_size()
		return {
			width = self._viewport_width,
			height = self._viewport_height
		}
	end

	function ViewportBorderImage:get_position()
		return {
			x = self._position_x,
			y = self._position_y
		}
	end

	function ViewportBorderImage:get_scale()
		return {
			x = self._scale_x,
			y = self._scale_y
		}
	end

-- ViewportBorderWhatHowImage
	ViewportBorderWhatHowImage = ViewportBorderWhatHowImage or class(ViewportBorderImage)
	ViewportBorderWhatHowImage.transition_time = 0.5
	ViewportBorderWhatHowImage.life_time = 9.0

	function ViewportBorderWhatHowImage:update(t, dt, viewport_aspect_ratio)
		if self._death_wanted then
			return true
		end

		-- Just incase the user resizes the resolution mid display.
		local force_recalculate = viewport_aspect_ratio ~= self._last_viewport_aspect_ratio

		if force_recalculate or not self._hidden_info then
			self._hidden_info = {}

			local container_width = 0
			local container_height = 0

			if self._width/self._height > viewport_aspect_ratio then
				container_width = self._width
				container_height = self._width / viewport_aspect_ratio
			else
				container_width = self._height * viewport_aspect_ratio
				container_height = self._height
			end

			self._hidden_info.scale_x = container_width / self._viewport_width
			self._hidden_info.scale_y = container_height / self._viewport_height

			local offset_x = (container_width - self._width) / 2
			local offset_y = (container_height - self._height) / 2

			self._hidden_info.position_x = -(((offset_x + self._viewport_x) / container_width) * self._hidden_info.scale_x)
			self._hidden_info.position_y = -(((offset_y + self._viewport_y) / container_height) * self._hidden_info.scale_y)
		end

		if force_recalculate or not self._shown_info then
			self._shown_info = {
				scale_x = 1,
				scale_y = 1,
				position_x = 0,
				position_y = 0
			}
		end

		if not self._start_t then
			self._start_t = t
			self._end_t = t + ViewportBorderWhatHowImage.life_time
		end

		local lerp_value = 1
		if t <= self._start_t + ViewportBorderWhatHowImage.transition_time  then
			lerp_value = (t - self._start_t) / ViewportBorderWhatHowImage.transition_time
		elseif t <= self._end_t and t > self._end_t - ViewportBorderWhatHowImage.transition_time then
			lerp_value = (self._end_t - t) / ViewportBorderWhatHowImage.transition_time
		elseif t > self._end_t then
			return true
		end

		self._scale_x = math.lerp(self._hidden_info.scale_x, self._shown_info.scale_x, lerp_value)
		self._scale_y = math.lerp(self._hidden_info.scale_y, self._shown_info.scale_y, lerp_value)

		self._position_x = math.lerp(self._hidden_info.position_x, self._shown_info.position_x, lerp_value)
		self._position_y = math.lerp(self._hidden_info.position_y, self._shown_info.position_y, lerp_value)

		self._last_viewport_aspect_ratio = viewport_aspect_ratio
	end

-- ViewportBorderDVD
	ViewportBorderDVD = ViewportBorderDVD or class(ViewportBorderImage)
	ViewportBorderDVD.scale = 1/4
	ViewportBorderDVD.speed = 500
	ViewportBorderDVD.start_angle = 45.0

	function ViewportBorderDVD:init()
		self._width = 1920
		self._height = 1080

		self._viewport_x = 0
		self._viewport_y = 0

		self._viewport_width = self._width * ViewportBorderDVD.scale
		self._viewport_height = self._height * ViewportBorderDVD.scale

		self._position_x = 0
		self._position_y = 0

		self._scale_x = 1
		self._scale_y = 1

		self._direction = {
			x = math.cos(ViewportBorderDVD.start_angle),
			y = math.sin(ViewportBorderDVD.start_angle)
		}

		self._death_wanted = false
	end

	function ViewportBorderDVD:update(t, dt, viewport_aspect_ratio)
		self._width = 1080 * viewport_aspect_ratio
		self._height = 1080

		self._viewport_width = (self._height * ViewportBorderDVD.scale) * 16/9
		self._viewport_height = self._height * ViewportBorderDVD.scale

		self._viewport_x = self._viewport_x + (dt * self._direction.x * ViewportBorderDVD.speed)
		self._viewport_y = self._viewport_y + (dt * self._direction.y * ViewportBorderDVD.speed)

		if self._viewport_x < 0 then
			self._viewport_x = 0
			self._direction.x = -self._direction.x
		elseif self._viewport_x + self._viewport_width > self._width then
			self._viewport_x = self._width - self._viewport_width
			self._direction.x = -self._direction.x
		end

		if self._viewport_y < 0 then
			self._viewport_y = 0
			self._direction.y = -self._direction.y
		elseif self._viewport_y + self._viewport_height > self._height then
			self._viewport_y = self._height - self._viewport_height
			self._direction.y = -self._direction.y
		end

		return self._death_wanted
	end

ViewportBorder = ViewportBorder or {}
ViewportBorder._workspace = nil

ViewportBorder._layers = {}

function ViewportBorder.add_layer(image_class, index)
	local index = index or (#ViewportBorder._layers + 1)

	table.insert(ViewportBorder._layers, index, {image_class, nil})
end

function ViewportBorder._update_layer_bitmaps(t, dt)
	local res = RenderSettings.resolution

	if not ViewportBorder._workspace then return end
	local panel = ViewportBorder._workspace:panel()

	local next_bitmap_corner_offset = { x = 0, y = 0 }
	local next_bitmap_resolution = RenderSettings.resolution

	local indexes_to_kill = {}

	for index = 1, #ViewportBorder._layers do
		table_index = #ViewportBorder._layers + 1 - index
		layer_info = ViewportBorder._layers[table_index]

		if layer_info[1]:update(t, dt, next_bitmap_resolution.x/next_bitmap_resolution.y) then
			table.insert(indexes_to_kill, table_index)

			if layer_info[2] then
				layer_info[2]:set_visible(false)
			end
		else
			local layer_size = layer_info[1]:get_size()
			local layer_aspect = layer_size.width / layer_size.height

			local scaled_layer_size = { width = 0, height = 0 }
			if next_bitmap_resolution.x/next_bitmap_resolution.y < layer_aspect then
				scaled_layer_size.width = next_bitmap_resolution.x
				scaled_layer_size.height = next_bitmap_resolution.x / layer_aspect
			else
				scaled_layer_size.height = next_bitmap_resolution.y
				scaled_layer_size.width = next_bitmap_resolution.y * layer_aspect
			end

			local layer_scale = layer_info[1]:get_scale()
			scaled_layer_size.width = scaled_layer_size.width * layer_scale.x
			scaled_layer_size.height = scaled_layer_size.height * layer_scale.y

			local layer_viewport_position = layer_info[1]:get_viewport_position()
			local x_percentage = layer_viewport_position.x/layer_size.width
			local y_percentage = layer_viewport_position.y/layer_size.height

			local layer_viewport_size = layer_info[1]:get_viewport_size()
			local height_percentage = layer_viewport_size.height/layer_size.height
			local width_percentage = layer_viewport_size.width/layer_size.width

			local layer_source = layer_info[1]:get_source()
			layer_info[2] = layer_info[2] or panel:bitmap({})

			local layer_position = layer_info[1]:get_position()
			local bitmap_lefttop = {
				x = next_bitmap_corner_offset.x + (layer_position.x * next_bitmap_resolution.x) + (((next_bitmap_resolution.x * layer_scale.x) - scaled_layer_size.width)/2),
				y = next_bitmap_corner_offset.y + (layer_position.y * next_bitmap_resolution.y) + (((next_bitmap_resolution.y * layer_scale.y) - scaled_layer_size.height)/2)
			}

			if layer_source then
				layer_info[2]:set_image(layer_source)
				layer_info[2]:set_layer(-index)

				layer_info[2]:set_size(
					scaled_layer_size.width,
					scaled_layer_size.height * layer_aspect
				)

				layer_info[2]:set_lefttop(
					bitmap_lefttop.x,
					bitmap_lefttop.y
				)

				layer_info[2]:set_visible(true)
			else
				layer_info[2]:set_visible(false)
			end

			next_bitmap_corner_offset = {
				x = bitmap_lefttop.x + (scaled_layer_size.width * x_percentage),
				y = bitmap_lefttop.y + (scaled_layer_size.height * y_percentage)
			}

			next_bitmap_resolution = {
				x = scaled_layer_size.width * width_percentage,
				y = scaled_layer_size.height * height_percentage
			}
		end
	end

	for _, table_index in pairs(indexes_to_kill) do
		table.remove(ViewportBorder._layers, table_index)
	end

	ViewportBorder.set_viewport_settings(
		next_bitmap_corner_offset.x/res.x, next_bitmap_corner_offset.y/res.y,
		next_bitmap_resolution.x/res.x, next_bitmap_resolution.y/res.y,
		next_bitmap_resolution.x/next_bitmap_resolution.y
	)
end

ViewportBorder._viewport = nil

function ViewportBorder.set_viewport(viewport)
	ViewportBorder._viewport = viewport
end

ViewportBorder._current_viewport_settings = {
	x = 0, y = 0, width = 1, height = 1, aspect_ratio = RenderSettings.resolution.x/RenderSettings.resolution.y
}

function ViewportBorder.set_viewport_settings(x, y, width, height, aspect_ratio)
	ViewportBorder._current_viewport_settings = {
		x = x, y = y, width = width, height = height, aspect_ratio = aspect_ratio
	}
end

function ViewportBorder._update_viewport()
	if ViewportBorder._viewport and alive(ViewportBorder._viewport._vp) then
		local settings = ViewportBorder._current_viewport_settings

		local viewport_scaling = managers.user:get_setting("viewport_scale") or 1

		local width = settings.width * viewport_scaling
		local height = settings.height * viewport_scaling

		local position_x = settings.x + ((settings.width - width) / 2)
		local position_y = settings.y + ((settings.height - height) / 2)

		ViewportBorder._viewport._vp:set_dimensions(position_x, position_y, width, height)
		ViewportBorder._viewport:camera():set_width_multiplier(CoreMath.width_mul(settings.aspect_ratio))
	end
end

function ViewportBorder._get_safe_rect_pixels()
	local width_mul = ViewportBorder._current_viewport_settings.width * RenderSettings.resolution.x
	local height_mul = ViewportBorder._current_viewport_settings.height * RenderSettings.resolution.y

	local safe_rect_scale = managers.viewport:get_safe_rect()
	local safe_rect_pixels = {
		x = math.round(safe_rect_scale.x * width_mul),
		y = math.round(safe_rect_scale.y * height_mul),
		width = math.round(safe_rect_scale.width * width_mul),
		height = math.round(safe_rect_scale.height * height_mul)
	}

	return safe_rect_pixels
end

ViewportBorder._world_to_screen_workspaces = {
	saferect = nil,
	workspace = nil
}

function ViewportBorder._update_workspaces()
	if not (managers.hud._saferect and alive(managers.hud._saferect) and managers.hud._workspace and alive(managers.hud._workspace)) then
		return
	end

	local safe_rect = ViewportBorder._get_safe_rect_pixels()
	local scaled_size = managers.gui_data:scaled_size()

	local res = RenderSettings.resolution
	local width_mul = ViewportBorder._current_viewport_settings.width * res.x
	local height_mul = ViewportBorder._current_viewport_settings.height * res.y

	local w = scaled_size.width
	local h = scaled_size.height
	local sh = math.min(safe_rect.height, safe_rect.width / (w / h))
	local sw = math.min(safe_rect.width, safe_rect.height * w / h)
	local x = (width_mul / 2 - sh * w / h / 2)
	local y = (height_mul / 2 - sw / (w / h) / 2)

	managers.hud._saferect:set_screen(
		w,
		h,
		x + (ViewportBorder._current_viewport_settings.x * res.x),
		y + (ViewportBorder._current_viewport_settings.y * res.y),
		sw
	)

	local original_safe_rect = managers.gui_data:_get_safe_rect_pixels()
	local original_sh = math.min(original_safe_rect.height, original_safe_rect.width / (w / h))
	local original_sw = math.min(original_safe_rect.width, original_safe_rect.height * w / h)
	local original_x = res.x / 2 - original_sh * w / h / 2
	local original_y = res.y / 2 - original_sw / (w / h) / 2

	ViewportBorder._world_to_screen_workspaces.saferect = ViewportBorder._world_to_screen_workspaces.saferect or managers.gui_data:create_saferect_workspace("screen", Overlay:gui())
	ViewportBorder._world_to_screen_workspaces.saferect:set_screen(
		w,
		h,
		original_x,
		original_y,
		original_sw
	)

	local gui_width, gui_height = managers.gui_data:get_base_res()

	sh = gui_width / ViewportBorder._current_viewport_settings.aspect_ratio
	h = math.max(gui_height, sh)
	w = math.max(gui_width, gui_width / (sh / h))

	managers.hud._workspace:set_screen(
		w,
		h,
		ViewportBorder._current_viewport_settings.x * res.x,
		ViewportBorder._current_viewport_settings.y * res.y,
		width_mul
	)

	ViewportBorder._world_to_screen_workspaces.workspace = ViewportBorder._world_to_screen_workspaces.workspace or managers.gui_data:create_fullscreen_workspace("screen", Overlay:gui())
	ViewportBorder._world_to_screen_workspaces.workspace:set_screen(
		w,
		h,
		0,
		0,
		res.x
	)

	ViewportBorder._workspace = ViewportBorder._workspace or managers.gui_data:create_fullscreen_workspace("screen", Overlay:gui())
	ViewportBorder._workspace:set_screen(
		res.x,
		res.y,
		0,
		0,
		res.x
	)
end

Hooks:PostHook(HUDManager, "update", "ViewportBorderUpdate", function(self, t, dt)
	ViewportBorder._update_layer_bitmaps(t, dt)

	ViewportBorder._update_viewport()
	ViewportBorder._update_workspaces()
end)

-- Floating stuff replacement.
	local wp_pos = Vector3()
	local wp_dir = Vector3()
	local wp_dir_normalized = Vector3()
	local wp_cam_forward = Vector3()
	local wp_onscreen_direction = Vector3()
	local wp_onscreen_target_pos = Vector3()

	function HUDManager:_update_waypoints(t, dt)
		local cam = managers.viewport:get_current_camera()

		if not cam then
			return
		end

		local cam_pos = managers.viewport:get_current_camera_position()
		local cam_rot = managers.viewport:get_current_camera_rotation()

		mrotation.y(cam_rot, wp_cam_forward)

		for id, data in pairs(self._hud.waypoints) do
			local panel = data.bitmap:parent()

			if data.state == "dirty" then
				-- Nothing
			end

			if data.state == "sneak_present" then
				data.current_position = Vector3(panel:center_x(), panel:center_y())

				data.bitmap:set_center_x(data.current_position.x)
				data.bitmap:set_center_y(data.current_position.y)

				data.slot = nil
				data.current_scale = 1
				data.state = "present_ended"
				data.text_alpha = 0.5
				data.in_timer = 0
				data.target_scale = 1

				if data.distance then
					data.distance:set_visible(true)
				end
			elseif data.state == "present" then
				data.current_position = Vector3(panel:center_x() + data.slot_x, panel:center_y() + panel:center_y() / 2)

				data.bitmap:set_center_x(data.current_position.x)
				data.bitmap:set_center_y(data.current_position.y)
				data.text:set_center_x(data.bitmap:center_x())
				data.text:set_top(data.bitmap:bottom())

				data.present_timer = data.present_timer - dt

				if data.present_timer <= 0 then
					data.slot = nil
					data.current_scale = 1
					data.state = "present_ended"
					data.text_alpha = 0.5
					data.in_timer = 0
					data.target_scale = 1

					if data.distance then
						data.distance:set_visible(true)
					end
				end
			else
				if data.text_alpha and data.text_alpha ~= 0 then
					data.text_alpha = math.clamp(data.text_alpha - dt, 0, 1)

					data.text:set_color(data.text:color():with_alpha(data.text_alpha))
				end

				data.position = data.unit and data.unit:position() or data.position

				mvector3.set(wp_pos, (ViewportBorder._world_to_screen_workspaces.saferect or self._saferect):world_to_screen(cam, data.position))
				mvector3.set(wp_dir, data.position)
				mvector3.subtract(wp_dir, cam_pos)
				mvector3.set(wp_dir_normalized, wp_dir)
				mvector3.normalize(wp_dir_normalized)

				local dot = mvector3.dot(wp_cam_forward, wp_dir_normalized)

				if dot < 0 or panel:outside(mvector3.x(wp_pos), mvector3.y(wp_pos)) then
					if data.state ~= "offscreen" then
						data.state = "offscreen"

						data.arrow:set_visible(true)
						data.bitmap:set_color(data.bitmap:color():with_alpha(0.75))

						data.off_timer = 0 - (1 - data.in_timer)
						data.target_scale = 0.75

						if data.distance then
							data.distance:set_visible(false)
						end

						if data.timer_gui then
							data.timer_gui:set_visible(false)
						end
					end

					local direction = wp_onscreen_direction
					local panel_center_x, panel_center_y = panel:center()

					mvector3.set_static(direction, wp_pos.x - panel_center_x, wp_pos.y - panel_center_y, 0)
					mvector3.normalize(direction)

					local distance = data.radius * tweak_data.scale.hud_crosshair_offset_multiplier
					local target_pos = wp_onscreen_target_pos

					mvector3.set_static(target_pos, panel_center_x + mvector3.x(direction) * distance, panel_center_y + mvector3.y(direction) * distance, 0)

					data.off_timer = math.clamp(data.off_timer + dt / data.move_speed, 0, 1)

					if data.off_timer ~= 1 then
						mvector3.set(data.current_position, math.bezier({
							data.current_position,
							data.current_position,
							target_pos,
							target_pos
						}, data.off_timer))

						data.current_scale = math.bezier({
							data.current_scale,
							data.current_scale,
							data.target_scale,
							data.target_scale
						}, data.off_timer)

						data.bitmap:set_size(data.size.x * data.current_scale, data.size.y * data.current_scale)
					else
						mvector3.set(data.current_position, target_pos)
					end

					data.bitmap:set_center(mvector3.x(data.current_position), mvector3.y(data.current_position))
					data.arrow:set_center(mvector3.x(data.current_position) + direction.x * 24, mvector3.y(data.current_position) + direction.y * 24)

					local angle = math.X:angle(direction) * math.sign(direction.y)

					data.arrow:set_rotation(angle)

					if data.text_alpha ~= 0 then
						data.text:set_center_x(data.bitmap:center_x())
						data.text:set_top(data.bitmap:bottom())
					end
				else
					if data.state == "offscreen" then
						data.state = "onscreen"

						data.arrow:set_visible(false)
						data.bitmap:set_color(data.bitmap:color():with_alpha(1))

						data.in_timer = 0 - (1 - data.off_timer)
						data.target_scale = 1

						if data.distance then
							data.distance:set_visible(true)
						end

						if data.timer_gui then
							data.timer_gui:set_visible(true)
						end
					end

					local alpha = 0.8

					if dot > 0.99 then
						alpha = math.clamp((1 - dot) / 0.01, 0.4, alpha)
					end

					if data.bitmap:color().alpha ~= alpha then
						data.bitmap:set_color(data.bitmap:color():with_alpha(alpha))

						if data.distance then
							data.distance:set_color(data.distance:color():with_alpha(alpha))
						end

						if data.timer_gui then
							data.timer_gui:set_color(data.bitmap:color():with_alpha(alpha))
						end
					end

					if data.in_timer ~= 1 then
						data.in_timer = math.clamp(data.in_timer + dt / data.move_speed, 0, 1)

						mvector3.set(data.current_position, math.bezier({
							data.current_position,
							data.current_position,
							wp_pos,
							wp_pos
						}, data.in_timer))

						data.current_scale = math.bezier({
							data.current_scale,
							data.current_scale,
							data.target_scale,
							data.target_scale
						}, data.in_timer)

						data.bitmap:set_size(data.size.x * data.current_scale, data.size.y * data.current_scale)
					else
						mvector3.set(data.current_position, wp_pos)
					end

					data.bitmap:set_center(mvector3.x(data.current_position), mvector3.y(data.current_position))

					if data.text_alpha ~= 0 then
						data.text:set_center_x(data.bitmap:center_x())
						data.text:set_top(data.bitmap:bottom())
					end

					if data.distance then
						local length = wp_dir:length()

						data.distance:set_text(string.format("%.0f", length / 100) .. "m")
						data.distance:set_center_x(data.bitmap:center_x())
						data.distance:set_top(data.bitmap:bottom())
					end
				end
			end

			if data.timer_gui then
				data.timer_gui:set_center_x(data.bitmap:center_x())
				data.timer_gui:set_bottom(data.bitmap:top())

				if data.pause_timer == 0 then
					data.timer = data.timer - dt
					local text = data.timer < 0 and "00" or (math.round(data.timer) < 10 and "0" or "") .. math.round(data.timer)

					data.timer_gui:set_text(text)
				end
			end
		end
	end

	local nl_w_pos = Vector3()
	local nl_pos = Vector3()
	local nl_dir = Vector3()
	local nl_dir_normalized = Vector3()
	local nl_cam_forward = Vector3()

	function HUDManager:_update_name_labels(t, dt)
		local cam = managers.viewport:get_current_camera()

		if not cam then
			return
		end

		local player = managers.player:local_player()
		local in_steelsight = false

		if alive(player) then
			in_steelsight = player:movement() and player:movement():current_state() and player:movement():current_state():in_steelsight() or false
		end

		local cam_pos = managers.viewport:get_current_camera_position()
		local cam_rot = managers.viewport:get_current_camera_rotation()

		mrotation.y(cam_rot, nl_cam_forward)

		local to_remove = {}
		local panel = nil

		for _, data in ipairs(self._hud.name_labels) do
			local label_panel = data.panel
			panel = panel or label_panel:parent()
			local pos = nil

			if data.movement then
				if not alive(data.movement._unit) then
					label_panel:set_visible(false)

					to_remove[data.id] = true
				else
					pos = data.movement:m_pos()

					mvector3.set(nl_w_pos, pos)
					mvector3.set_z(nl_w_pos, mvector3.z(data.movement:m_head_pos()) + 30)
				end
			elseif data.vehicle then
				if not alive(data.vehicle) then
					return
				end

				pos = data.vehicle:position()

				mvector3.set(nl_w_pos, pos)
				mvector3.set_z(nl_w_pos, pos.z + data.vehicle:vehicle_driving().hud_label_offset)
			end

			if pos then
				mvector3.set(nl_pos, (ViewportBorder._world_to_screen_workspaces.workspace or self._workspace):world_to_screen(cam, nl_w_pos))
				mvector3.set(nl_dir, nl_w_pos)
				mvector3.subtract(nl_dir, cam_pos)
				mvector3.set(nl_dir_normalized, nl_dir)
				mvector3.normalize(nl_dir_normalized)

				local dot = mvector3.dot(nl_cam_forward, nl_dir_normalized)

				if dot < 0 or panel:outside(mvector3.x(nl_pos), mvector3.y(nl_pos)) then
					if label_panel:visible() then
						label_panel:set_visible(false)
					end
				else
					label_panel:set_alpha(in_steelsight and math.clamp((1 - dot) * 100, 0, 1) or 1)
					label_panel:set_visible(true)

					if mvector3.distance_sq(cam_pos, nl_w_pos) < 250000 then
						label_panel:set_visible(true)
					elseif dot > 0.925 then
						-- Nothing
					end
				end

				if data.movement then
					if data.movement.current_state_name and data.movement:current_state_name() == "driving" then
						label_panel:set_visible(false)
					elseif data.movement.vehicle_seat and data.movement.vehicle_seat.occupant ~= nil then
						label_panel:set_visible(false)
					end
				end

				local offset = data.panel:child("cheater"):h() / 2

				if label_panel:visible() then
					label_panel:set_center(nl_pos.x, nl_pos.y - offset)
				end
			end
		end

		for id, _ in pairs(to_remove) do
			self:_remove_name_label(id)
		end
	end